'''
@Time    : 2022.04.23
@Author  : wuruidong
@Email   : wuruidong@hotmail.com
@FileName: LeNet_onnx.py
@Software: python pytorch=1.6.0 ppq=0.6.3 onnx=1.8.1
@Cnblogs : https://www.cnblogs.com/ruidongwu
'''
import torch as tf
import torch.nn as nn
import torch.nn. functional as F
import torch.optim as optim
from torch.autograd import Variable
from torchvision import datasets, transforms

# convenience class to keep track of averages
class Metric(object):
    def __init__(self, name):
        self.name = name
        self.sum = tf.tensor(0.)
        self.n = tf.tensor(0.)
    def update(self, val):
        self.sum += val.cpu()
        self.n += 1
    @property
    def avg(self):
        return self.sum / self.n

class LeNet(tf.nn.Module):
    def __init__(self):
        super(LeNet, self).__init__()
        self.conv1 = nn.Conv2d(1, 4, 3, 1)
        self.bn1 = nn.BatchNorm2d(4)
        self.relu1 = nn.ReLU(inplace=False)  # <== Module, not Function!
        self.pool1 = nn.MaxPool2d(2)

        self.conv2 = nn.Conv2d(4, 8, 3, 1)
        self.bn2 = nn.BatchNorm2d(8)
        self.relu2 = nn.ReLU(inplace=False)  # <== Module, not Function!
        #self.pad = nn.ZeroPad2d(padding=(1,0,1,0))
        self.pool2 = nn.MaxPool2d(2, padding=1)

        self.conv3 = nn.Conv2d(8, 16, 3, 1)
        self.bn3 = nn.BatchNorm2d(16)
        self.relu3 = nn.ReLU(inplace=False)  # <== Module, not Function!
        self.pool3 = nn.MaxPool2d(2)

        self.fc1 = nn.Linear(64, 10)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu1(x)  # <== Module, not Function!
        x = self.pool1(x)

        x = self.conv2(x)
        x = self.bn2(x)
        x = self.relu2(x)  # <== Module, not Function!
        #x = self.pad(x)
        x = self.pool2(x)

        x = self.conv3(x)
        x = self.bn3(x)
        x = self.relu3(x)  # <== Module, not Function!
        x = self.pool3(x)

        x = tf.flatten(x, 1)
        x = self.fc1(x)
        #output = F.log_softmax(x, dim=1)  # <== the softmax operation does not need to be quantized, we can keep it as it is
        output = x
        return output

def test(model, device, test_loader, integer=False, verbose=False):
    model.eval()
    test_loss = 0
    correct = 0
    test_acc = Metric('test_acc')

    with tf.no_grad():
        for data, target in test_loader:
            if integer:      # <== this will be useful when we get to the
                data *= 255  #     IntegerDeployable stage
            data, target = data.to(device), target.to(device)
            output = model(data)
            test_loss += F.nll_loss(output, target, reduction='sum').item() # sum up batch loss
            pred = output.argmax(dim=1, keepdim=True) # get the index of the max log-probability
            correct += pred.eq(target.view_as(pred)).sum().item()
            test_acc.update((pred == target.view_as(pred)).float().mean())

    test_loss /= len(test_loader.dataset)
    return test_acc.avg.item() * 100.

device = tf.device("cuda" if tf.cuda.is_available() else "cpu")
print('current device is', device)

train_loader = tf.utils.data.DataLoader(
    datasets.MNIST('../data', train=True, download=False, transform=transforms.Compose([
        transforms.ToTensor()
    ])),
    batch_size=128, shuffle=True
)
test_loader = tf.utils.data.DataLoader(
    datasets.MNIST('../data', train=False, transform=transforms.Compose([
        transforms.ToTensor()
    ])),
    batch_size=128, shuffle=False
)


model = LeNet()
if tf.cuda.is_available():
    model = model.cuda()

criterion = nn.CrossEntropyLoss()
optimizer = optim.SGD(model.parameters(), lr=1e-2)

losses = []
acces = []
eval_losses = []
eval_acces = []

for e in range(10):
    train_loss = 0
    train_acc = 0
    model.train()
    for im, label in train_loader:
        im = Variable(im)*255
        label = Variable(label)
        if tf.cuda.is_available():
            im = im.cuda()
            label = label.cuda()
        out = model(im)
        loss = criterion(out, label)

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        train_loss += loss.item()

        _, pred = out.max(1)
        num_correct = (pred==label).sum().item()

        acc = num_correct/im.shape[0]
        train_acc += acc

    losses.append(train_loss/len(train_loader))
    acces.append(train_acc/len(train_loader))

    eval_loss = 0
    eval_acc = 0
    model.eval()

    acc = test(model, device, test_loader, integer=True)
    print(e, "FullPrecision accuracy: %.02f%%" % acc)


tf.save(model.state_dict(), 'LeNet.pth')


dummy_input = tf.randn(16, 1, 28, 28, device='cuda')
input_names = [ "input" ]
output_names = [ "output" ]

tf.onnx.export(model, dummy_input, "LeNet.onnx", verbose=True, input_names=input_names, output_names=output_names)
print('Application is over!')
